from typing import List
from dataclasses import dataclass, field


import spotipy
from spotipy.exceptions import SpotifyException
from spotipy.oauth2 import SpotifyOAuth


# import spotidl.exceptions as exceptions
# import spotidl.utils as utils

from spotidl import exceptions, utils


# initializing the spotify api connection
# the OAuth object automatically reads valid env. variables so we don't need to
# manually assign them using `os.environ.get(<env_var name>)`
try:
    sp = spotipy.Spotify(auth_manager=SpotifyOAuth())

except spotipy.oauth2.SpotifyOauthError as ex:
    raise exceptions.EnvVariablesError(ex)


# defining structure for the song data we are going to get
@dataclass
class SpotifySong:
    """
    Stores data for the song fetched from Spotify.
    """

    name: str = "Unknown Song"
    # this generates a list with an unknown artist default
    artists: list = field(default_factory=lambda: ["Unknown Artist"])
    album_name: str = "Unknown Album"
    disc_number: int = 1
    track_number: int = 1
    cover_url: str = ""

    def __str__(self):
        return utils.make_song_title(self.artists, self.name, ", ")


def get_song_data(link: str) -> SpotifySong:
    """
    Gets relevant song details for the given Spotify song link.
    These are then passed onto other functions for further processing.
    """

    try:
        query = sp.track(link)

    except SpotifyException as exception:
        # wrapping the Spotify Exception
        # still unsure whether this is the correct approach
        raise exceptions.NoDataReceivedError(exception)

    else:
        # there can be multiple results,
        # iterating over the list and extracting data
        artists = []
        for artist in query["artists"]:
            artists.append(artist["name"])

        song = new_song(query, type_="song")

        return song


# this will serve as the common point for all entry types
# (individual song link, album link, etc.)
def new_song(query: dict, type_: str, **album_details) -> SpotifySong:
    """
    Makes a new SpotifySong given a raw "track" type item received from Spotify API.
    The track queried using different methods returns slightly modified data.
    """

    artists = []
    for artist in query["artists"]:
        artists.append(artist["name"])

    name = utils.correct_name(query["name"])

    if type_ == "song":
        album_name = utils.correct_name(query["album"]["name"])
        # "song" refers to the case where the user enters a song link; we need
        # to fetch data for just a single song
        song = SpotifySong(
            name=name,
            artists=artists,
            album_name=album_name,
            disc_number=query["disc_number"],
            track_number=query["track_number"],
            # the 1st link contains a link to the album art
            # of size 640 x 640 pixels
            cover_url=query["album"]["images"][0]["url"],
        )

    elif type_ == "album":
        song = SpotifySong(
            name=name,
            artists=artists,
            album_name=album_details["album_name"],
            disc_number=query["disc_number"],
            track_number=query["track_number"],
            cover_url=album_details["album_cover_url"],
        )

    return song


def get_album_data(link: str) -> tuple:
    """
    Gets album data for the given Spotify album link.
    It is then passed onto other functions for further processing.
    """

    # since our main data(all the relevant metadata, song URL) is associated
    # with the spotify song(referred to as "Track" in the spotify api), we
    # will need to make a list of all the songs in the album and then pass on
    # that list onto other functions.
    try:
        query = sp.album(link)

    except SpotifyException as exception:
        raise exceptions.NoDataReceivedError(exception)

    else:
        album: List[SpotifySong] = []
        album_name = utils.correct_name(query["name"])

        for track in query["tracks"]["items"]:
            # since this time we are directly receiving data that we otherwise
            # extracted from get_song_data using a link entered by the user, we
            # will need to generate a SpotifySong object another way

            song = new_song(
                track,
                type_="album",
                album_name=album_name,
                album_cover_url=query["images"][0]["url"],
            )

            album.append(song)

        # returning album name since we'll need it when making the album folder to
        # use as the save directory
        return (album_name, album)


def get_playlist_data(link: str) -> tuple:
    """
    Gets playlist data for the given Spotify playlist link.
    It is then passed onto other functions for further processing.
    """

    # we need playlist id to fetch its information
    playlist_id = utils.get_playlist_id(link)

    try:
        query = sp.playlist_tracks(playlist_id)
        playlist_name = sp.playlist(playlist_id)["name"]

    except SpotifyException as exception:
        raise exceptions.NoDataReceivedError(exception)

    else:
        # can fetch a 100 tracks at a time
        # the "next" dictionary key gets the next batch, if the no. of results
        # exceeds 100
        tracks = query["items"]

        while query["next"]:
            query = sp.next(query)
            tracks.extend(query["items"])

        # now, to extract data from each entry in the list and get a SpotifySong object
        playlist_songs: List[SpotifySong] = []

        for track in tracks:
            song = new_song(track["track"], type_="song")

            playlist_songs.append(song)

        return playlist_name, playlist_songs



from dataclasses import dataclass

from yt_dlp import YoutubeDL
import yt_dlp

from spotidl.spotify import SpotifySong
from spotidl.utils import make_song_title, check_file


@dataclass
class YoutubeSong:
    """
    Umbrella for the song data extracted from Youtube.
    """

    id: str
    title: str
    video_url: str


def get_config(user_params: dict, song: SpotifySong) -> dict:
    """
    Prepares the parameters that need to be passed onto the YoutubeDL object.
    """

    downloader_params = {
        "postprocessors": [
            {
                "key": "FFmpegExtractAudio",
                "preferredcodec": user_params["codec"],
                "preferredquality": user_params["quality"],
            }
        ],
        "outtmpl": f"{make_song_title(song.artists, song.name, ', ')}.%(ext)s",
        # "outtmpl": "%(artist)s-%(title)s.ext",
        "quiet": user_params["quiet"],
        "format": "bestaudio/best",
        "dynamic_mpd": False,
        "noplaylist": True,
        "prefer_ffmpeg": True,
    }

    return downloader_params


def get_downloader(params: dict):
    """
    Initiates the YoutubeDL class with the configured parameters.
    """

    return YoutubeDL(params=params)


def fetch_source(ydl: YoutubeDL, song: SpotifySong) -> YoutubeSong:
    """
    Fetch appropriate source for the song from Youtube using the given details.
    """

    try:
        # adding "audio" to avoid 'official music videos' and similar types
        song_title = make_song_title(song.artists, song.name, ", ") + " audio"

        search: dict = ydl.extract_info(f"ytsearch:{song_title}", download=False)

        # extracting the first entry from the nested dict
        yt_info = search["entries"][0]

        # we are unable to find the song
        if song.name not in yt_info["title"]:
            print("Couldn't find the apt audio source with that name, retrying...")

            # retrying the search but with album name added
            song_title = (
                make_song_title(song.artists, song.name, ", ")
                + f" {song.album_name} audio"
            )

            search: dict = ydl.extract_info(f"ytsearch:{song_title}", download=False)

            yt_info = search["entries"][0]

            # now, if we are still getting the wrong result,
            # we should avoid the download
            return None

    except yt_dlp.DownloadError as ex:
        print("Error when trying to get audio source from YT: ", ex)
        return

    else:
        yt_song = YoutubeSong(
            id=yt_info["id"],
            title=yt_info["title"],
            video_url=yt_info["webpage_url"],
        )

        return yt_song


def download_song(ydl: YoutubeDL, link: str, count: int = 0) -> bool:
    """
    Downloads the song given its source link and the YouTube downloader object.
    """

    while count < 3:
        try:
            # attempts to download the song using the best matched
            # youtube source link
            count += 1
            ydl.download(link)

        except yt_dlp.DownloadError as ex:
            if count >= 3:
                print("\nDownload failed: ", ex)
                return False

            print("\nDownload failed! Retrying...")
            # call the function again but increase the count figure
            download_song(ydl, link, count=count + 1)

        else:
            return True


def controller(user_params: dict, song: SpotifySong, file_name: str) -> bool:
    """
    Handles the flow of the download process for the given song.
    Initiates the configuration as per the user-defined parameters and chains
    the rest of functions together.
    """

    # check if song has already been downloaded before at some point;
    # only proceed with download if it doesn't
    if check_file(file_name):
        print(f"\n{file_name} already exists! Skipping download...\n")
        # False will ensure that we don't attempt to re-write metadata again
        return False

    # user parameters are used in the downloader parameters dictionary
    # the downloader_params dict is then passed onto the YoutubeDL object
    # when generating its instance.
    downloader_params = get_config(user_params, song)
    ydl = get_downloader(downloader_params)

    print(f"Starting '{song}' song download...\n")
    yt_song = fetch_source(ydl, song)

    if not yt_song:
        print("Couldn't find audio source for the song, skipping...\n")
        return False

    if not download_song(ydl, yt_song.video_url):
        # we already print the download failed statement when encountering
        # the exception in the function itself, so just return False here
        return False

    print(f"\nDownload for song '{song}' completed. Enjoy!")
    return True



import mutagen
import mutagen.easyid3
import mutagen.id3
import mutagen.flac

from spotidl.spotify import SpotifySong
from spotidl.utils import download_album_art


def add_metadata_mp3(file_name: str, song: SpotifySong, album_art_path: str):
    """
    Adds metadata to a MP3 file given it's path.
    File name must contain file extension as well.
    """

    # writing textual metadata to the file
    try:
        meta = mutagen.easyid3.EasyID3(file_name)

    except mutagen.id3.ID3NoHeaderError:
        meta = mutagen.File(file_name, easy=True)
        meta.add_tags()

    meta["title"] = song.name
    #  ID3v2.3 spec: the delimiter is a "/" for multiple entries.
    meta["artist"] = "/".join(song.artists)
    meta["album"] = song.album_name
    meta["tracknumber"] = str(song.track_number)
    meta["discnumber"] = str(song.disc_number)

    meta.save(v2_version=3)

    # here we write the downloaded album art image to our downloaded song.
    if album_art_path:
        meta = mutagen.id3.ID3(file_name)

        with open(album_art_path, "rb") as pic:
            meta["APIC"] = mutagen.id3.APIC(
                # type 3 is for the front cover of a song
                encoding=3,
                mime="image/jpeg",
                type=3,
                data=pic.read(),
            )

        # setting the v2 version to 3 is key- nothing will work without it
        meta.save(v2_version=3)


def add_metadata_flac(file_name: str, song: SpotifySong, album_art_path: str):
    """
    Adds metadata to a FLAC file given it's path.
    File name must contain file extension as well.
    """

    meta = mutagen.flac.FLAC(file_name)

    try:
        meta.add_tags()

    except mutagen.flac.FLACVorbisError:
        # if this error occurs, the vorbis comment block already exists
        meta.delete()

    meta["title"] = song.name
    meta["artist"] = "/".join(song.artists)
    meta["album"] = song.album_name
    meta["tracknumber"] = str(song.track_number)
    meta["discnumber"] = str(song.disc_number)

    # now, to write the album art to the file
    if album_art_path:
        pic = mutagen.flac.Picture()

        with open(album_art_path, "rb") as pic:
            pic.data = pic.read()

        meta.add_picture(pic)

        meta.save()


# setting default directory to be the current folder(represented as ".")
def controller(file_name: str, song: SpotifySong, codec: str, directory: str = "."):
    """
    Handles the metadata writing process flow.
    """

    album_art_path = download_album_art(
        path=directory, link=song.cover_url, title=song.album_name
    )

    if not album_art_path:
        print("Album art couldn't be retrieved.")

    print("\nWriting metadata...")

    if codec == "flac":
        add_metadata_flac(file_name, song, album_art_path)

    elif codec == "mp3":
        add_metadata_mp3(file_name, song, album_art_path)

    print("\nFinished writing metadata!")



# contains the custom exceptions to be used when attempting to enforce
# certain aspects in the application

# not entering anything in the exceptions since all I want to do is to
# convey to the end user that something is wrong.
# a custom exception allows me to do that, with a related name and a custom
# description to explain the same in simple terms


class LinkError(Exception):
    """
    Raised when the provided link is incorrect.
    """

    def __init__(self, message="Invalid Spotify link entered!"):
        self.message = message
        super().__init__(self.message)


class NoDataReceivedError(Exception):
    """
    Raised when no data received from the library service.
    """

    def __init__(self, message="No data received from the service."):
        self.message = message
        super().__init__(self.message)


class EnvVariablesError(Exception):
    """
    Raised when environment variables aren't configured.
    """

    def __init__(self, message="Environment variables aren't configured properly!"):
        self.message = message
        super().__init__(self.message)


class FFmpegNotInstalledError(Exception):
    """
    Raised when FFmpeg is not found in the user's PATH.
    """

    def __init__(self, message="FFMpeg is not installed!"):
        self.message = message
        super().__init__(self.message)



import os
import subprocess
import platform
import re
from urllib.request import urlretrieve

from spotidl import config, exceptions


default_save_dir = os.getcwd() + "/dl"


def load_env_vars() -> dict:
    """
    Loads the environment variables into a dictionary.
    """

    client_id = os.getenv("SPOTIPY_CLIENT_ID")
    client_secret = os.getenv("SPOTIPY_CLIENT_SECRET")
    redirect_uri = os.getenv("SPOTIPY_REDIRECT_URI")

    env_vars = {
        "id": client_id,
        "secret": client_secret,
        "redirect_uri": redirect_uri,
    }

    return env_vars


def check_env_vars(env_vars: dict):
    """
    Run a barebones check to ensure that the three needed environment variables
    are not blank.
    """

    if not env_vars["id"]:
        raise exceptions.EnvVariablesError("SPOTIPY_CLIENT_ID not configured!")

    if not env_vars["secret"]:
        raise exceptions.EnvVariablesError("SPOTIPY_CLIENT_SECRET not configured!")

    if not env_vars["redirect_uri"]:
        raise exceptions.EnvVariablesError("SPOTIPY_REDIRECT_URI not configured!")


def make_dir(path: str) -> bool:
    """
    Creates a directory given a path.
    """

    try:
        os.makedirs(path, exist_ok=True)

    except OSError as exception:
        print("Error when attempting to make directory: ", exception)
        return False

    else:
        return True


# two helper functions to avoid having to import the os package in each module
# individually wherever we need to make use of file or directory checks
def check_dir(path: str) -> bool:
    """
    Checks whether a directory exists or not.
    """

    return os.path.isdir(path)


def check_file(path: str) -> bool:
    """
    Check whether a file exists or not.
    """

    return os.path.isfile(path)


def directory_maker(path: str):
    """
    Checks for the existence of a directory given the path
    and makes one if it doesn't exist.
    """

    if not check_dir(path):
        if make_dir(path):
            _, dir_name = os.path.split(path)
            print(f"Successfully created '{dir_name}' directory.")


def check_spotify_link(link: str, patterns_list: list) -> bool:
    """
    Handles all checks needed to vet user-entered Spotify links.
    """

    is_match = False

    # patterns_list contains a list of regex patterns for Spotify URLs
    for pattern in patterns_list:
        if re.search(pattern=pattern, string=link):
            is_match = True

    return is_match


def make_song_title(artists: list, name: str, delim: str) -> str:
    """
    Generates a song title by joining the song title and artist names.
    Artist names given in list format are split using the given delimiter.
    """

    return f"{delim.join(artists)} - {name}"


def download_album_art(
    path: str, link: str, title: str, extension: str = "jpeg"
) -> str:
    """
    Downloads album art- in a folder at the given path- to be embedded into songs.
    """

    if not link:
        return ""

    # make a folder in the given path to store album art
    folder = "/album-art"
    full_path = path + folder
    directory_maker(full_path)

    file = full_path + f"/{title}.{extension}"

    if not check_file(file):
        # urlretrieve downloads the resource from the given link and writes it
        # to the given file
        urlretrieve(link, file)

    return file


def check_cli_args(codec: str, bitrate: str, link: str) -> bool:
    """
    Corrects audio-related arguments if they are incorrect and finally calls
    for Spotify link checks.
    """

    # adding checks to ensure argument validity
    if codec not in config.audio_formats:
        print("Invalid codec entered! Using default value.")
        codec = "mp3"

    if bitrate not in config.audio_bitrates:
        print("Invalid bitrate entered! Using default value.")
        bitrate = "320"

    # if someone enters "" for example, after typing spotidl, it counts as an
    # arg for the parser but not for us
    if not link:
        raise exceptions.LinkError("Spotify link needed to proceed!")

    # check whether the provided link is authentic
    is_match = check_spotify_link(link, config.spotify_link_patterns)

    return is_match


def get_link_type(link: str) -> str:
    """
    Analyzes the given link and returns the type of download type required.
    (track for individual song, album for an album URL and playlist if given
    a the Spotify link of a playlist).
    """

    # a spotify link is in the form of: open.spotify.com/<type>/<id>
    # we only need the type part
    link_type = link.split("/")

    return link_type[len(link_type) - 2]


def correct_name(query: str) -> str:
    """
    Checks for illegal characters in file or folder name and corrects them
    if needed.
    """

    query = [char if char not in config.illegal_chars else "#" for char in query]
    query = "".join(query)

    return query


# since we cannot fetch any playlist data without a playlist id, this will be
# of paramount importance
def get_playlist_id(link: str) -> str:
    """
    Returns a playlist ID given a Spotify playlist URL.
    """

    # first remove the general URL part
    data = link.split("/")[-1]

    # we can safely return the last part if there's no question mark in it,
    # otherwise we need to split it again to remove the "si" code
    return data.split("?")[0] if "?" in data else data[-1]


# adding a check to see whether ffmpeg is installed since otherwise
# the user gets barraged with a whole lot of tracebacks
# this should simplify that
def check_ffmpeg_installed() -> bool:
    """
    Checks whether FFmpeg is installed or not.
    """

    os_platform = platform.system()

    try:
        if os_platform == "Windows":
            subprocess.check_call(["where", "ffmpeg"])

        else:
            subprocess.check_call(["which", "ffmpeg"])

    # this is an error returned when the platform can't be recognized,
    # from what I've seen
    # like, if the windows `where` command is run on a linux system, it'll pop
    # this exception
    except FileNotFoundError:
        return False

    return True



import argparse
import os

import dotenv


from spotidl import utils, spotify, downloader, config, exceptions, metadata
from . import __version__

# loading .env vars
dotenv.load_dotenv()


def cli_args() -> argparse.Namespace:
    """
    Contains and parses all command line arguments for the application.
    """

    parser = argparse.ArgumentParser(
        prog="spotidl",
        description="spotidl: download songs, albums and playlists using Spotify links",
    )
    parser.add_argument("link", help="Spotify song link to download")

    # important argument(s)
    parser.add_argument(
        "-d",
        "--dir",
        default=utils.default_save_dir,
        help="Save directory(is created if doesn't exist)",
    )

    # audio-related arguments
    # quiet is a 'stored' argument, means we don't have to enter anything
    # after calling "-q/--quiet", it defaults to True if called else False
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Makes the downloader non-verbose/quiet",
    )

    parser.add_argument(
        "-c",
        "--codec",
        default="mp3",
        help=f"Audio format to download file as. List of available formats: {config.audio_formats}",
    )

    parser.add_argument(
        "-b",
        "--bitrate",
        default="320",
        help=f"Audio quality of the file. List of available qualities: {config.audio_bitrates}",
    )

    # misc. arguments
    # tells the app version by reading it from the poetry toml file
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        # setting the version action's code will allow us to use this argument
        # during testing directly, instead of having to manually update the
        # tests each time
        version=__version__,
        help="Displays the current app version",
    )

    return parser.parse_args()


def prerun_checks(args: argparse.Namespace):
    """
    Performs all the necessary pre-run checks.
    """

    # we can't download and convert songs using the yt-dlp library w/o ffmpeg
    if not utils.check_ffmpeg_installed():
        raise exceptions.FFmpegNotInstalledError()

    #  perform necessary cli-argument validity checks
    if not utils.check_cli_args(args.codec, args.bitrate, args.link):
        raise exceptions.LinkError()

    # loading env vars
    env_vars = utils.load_env_vars()
    # checks to see whether required environment variables have been added
    # or not
    utils.check_env_vars(env_vars)


def controller():
    """
    Controls the flow of the program execution.
    """

    args = cli_args()
    prerun_checks(args)

    # i believe getting the link type should be separated from just checking
    # the validity of the link and the audio-related args like codec and bitrate
    # that are covered in the prerun_checks func
    link_type = utils.get_link_type(args.link)

    if not link_type in config.spotify_link_types:
        raise exceptions.LinkError("Invalid Spotify link type entered!")

    # make the specified dir. if it doesn't exist and open it to store files
    utils.directory_maker(args.dir)
    os.chdir(args.dir)

    # grouping all youtube-dl required arguments together before passing them
    # as the controller func parameters
    user_params = {
        "codec": args.codec,
        "quality": args.bitrate,
        "quiet": args.quiet,
        "dir": args.dir,
    }

    # replacing match-case with if else for comaptibility's sake
    if link_type == "track":
        song_download_controller(args.link, user_params)

    elif link_type == "album":
        album_download_controller(args.link, user_params)

    elif link_type == "playlist":
        playlist_download_controller(args.link, user_params)


def song_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download an individual song.
    """

    # gets the SpotifySong dataclass object to be used for everything else in the func
    song = spotify.get_song_data(link)

    # create the file name to be used when writing metadata
    file_name = f"{utils.make_song_title(song.artists, song.name, ', ')}.\
{user_params['codec']}"

    # use the youtube controller to download the song
    # we also send the filename here since the download controller checks
    # to ensure that the song hasn't been downloaded before.
    is_downloaded = downloader.controller(user_params, song, file_name)

    if is_downloaded:
        # write metadata to the downloaded file
        metadata.controller(
            file_name, song, codec=user_params["codec"], directory=user_params["dir"]
        )


def album_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download a complete album.
    """

    # get album information
    album_name, songs = spotify.get_album_data(link)
    save_dir = "./" + album_name

    # make a directory to store the album
    utils.directory_maker(save_dir)
    os.chdir(save_dir)

    print(f"Starting '{album_name}' album download...\n")

    for song in songs:
        file_name = f"{utils.make_song_title(song.artists, song.name, ', ')}.\
{user_params['codec']}"

        is_downloaded = downloader.controller(user_params, song, file_name)

        if is_downloaded:
            # write metadata to the downloaded file
            metadata.controller(file_name, song, codec=user_params["codec"])

    print(f"\nDownload for album '{album_name}' completed. Enjoy!")


def playlist_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download a complete playlist.
    """

    # get playlist information
    playlist_name, songs = spotify.get_playlist_data(link)
    save_dir = "./" + playlist_name

    # make a directory to store the playlist
    utils.directory_maker(save_dir)
    os.chdir(save_dir)

    print(f"Starting '{playlist_name}' playlist download...\n")

    for song in songs:
        file_name = f"{utils.make_song_title(song.artists, song.name, delim=', ')}.\
{user_params['codec']}"

        is_downloaded = downloader.controller(user_params, song, file_name)

        if is_downloaded:
            # write metadata to the downloaded file
            metadata.controller(file_name, song, codec=user_params["codec"])

    print(f"\nDownload for playlist '{playlist_name}' completed. Enjoy!")
